/*
   algebraic.cc -- handle higher order surfaces.

   (c) 1993, 1996 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>
 */


#include "object.hh"
#include "algebraic.hh"
#include "poly.hh"
#include "polyins.hh"
#include "ibuf.hh"
#include "proto.hh"
#include "extern.hh"
#include "vray.hh"
#include "stats.hh"

/// the polystack with polynomials
struct poly_evaluator {
    svec<Poly> polystack;

    /// execute one instruction
    void exec(const poly_instruction * instr);


    /// do the calculation
    void calc_all(const svec<poly_instruction> &instr);

    /// put the solution into argument
    void solution(Poly &);

    /// print the stack
    void print();

    ///   current ray.pos 
    vector raypos;

    /// ray.dir 
    vector raydir;  

    poly_evaluator(vector p,  vector d) {
	raypos=p;
	raydir=d;
    }

    /// generate poly for one algebraic. 
    void operator()(Poly &d,const svec<poly_instruction> &sv) {
	calc_all(sv);
	solution(d);
    }
	
};


/* dump stack */
 void
poly_evaluator::print()
{
    #ifdef DEBUG
    dprintf("\n----\n");
    for (    int     i = 0; i < polystack.sz(); i++)
    {
        dprintf("%d : ", i);
        polystack[i].print();
    }
    #endif
}

void
poly_evaluator::exec(const poly_instruction * instr)
{
    switch (instr->op)
    {
    case 'f':                   /* push a constant on the stack */
	polystack.add(Poly(instr->factor));
        return;

    case 'x':                   /* push X on stack */
    case 'y':                   /* push  Y on stack */	
    case 'z': {                 /* and Z */	
	axes a = axes((instr->op) -'x');
	polystack.add(Poly(raypos[a],raydir[a]));
	return;
    }

    }

    Poly  &curp  = polystack.last( );
    if (instr->op==  'n'){                   /* negate */
        curp.negate(curp);
        return;
    }

    Poly  &destp = polystack.last(1);
    
    switch (instr->op){
    case '-':                   /* substract */
        destp.subtract(destp,curp);
        polystack.pop();
        break;

    case '+':                   /* add */	
        destp.add(destp,curp);
        polystack.pop();
        break;

    case '*':                   /* multiply */
	destp.multiply(destp,curp);
        polystack.pop();
        break;
	
    case '^':                   /* power */
	{
	    int exponent = int(polystack.last().coef[0]);
	    destp.power(exponent, destp);	
	    polystack.pop();	    
	    break;
	}
    }
}


void
poly_evaluator::calc_all(svec< poly_instruction> const & instr)
{
    /* init stack */
    polystack.set_size(instr.sz());
    polystack.clear();
    
    /* execute poly instructions to form polynomial */
    for (int i = 0; i < instr.sz(); i++)
        exec(&instr[ i]);
}

void
poly_evaluator::solution(Poly &dst)
{
    assert(polystack.sz() == 1);
    dst = polystack[0];
}
/****************************************************************/

/// evaluate the gradient for an algebraic
struct normal_evaluator {
    svec<flt> dpolystack;
    
    vector normal_loc;
    int diff_to;

    void print();
    normal_evaluator(vector v) {normal_loc=v;}
    vector operator()(const svec<poly_instruction> & instr) {
	vector n;
	for (diff_to  = 0; diff_to < 3; diff_to++) {
	    n[diff_to] =calc_all(instr); 
	}

	return n;
    }
    flt calc_all(svec<poly_instruction> const & instr);
    /// handle one instruction, 
    void exec(poly_instruction * instr);
};


void
normal_evaluator ::exec(poly_instruction * instr)
{
    flt     p,
            Dp;

    switch (instr->op)
    {
    case 'x':
    case 'y':
    case 'z': {
	axes i = axes(instr->op - 'x');
        dpolystack.add( normal_loc[i]);
	dpolystack.add( diff_to == i);
        break;
    }
    case 'f':
	dpolystack.add(instr->factor);
	dpolystack.add(0);
        break;

    case '+':
    case '-':
	
	p=dpolystack.last(1);
	Dp=dpolystack .last();
	dpolystack .	pop(2);


        if(instr->op=='+') {
	    dpolystack.last()+=Dp;
	    /* (p1+p2)' = p1' + p2' */
	    dpolystack.last(1)+=p;
	} else {
	    dpolystack.last()-=Dp;
	    /* (p1+p2)' = p1' + p2' */
	    dpolystack.last(1)-=p;
	}
        break;
    
    case '*':
    case '^':
    {

	flt*fp=&dpolystack.last(3);	
	flt p1=*fp++,		// foei! dependant on implementation svec.
	    Dp1=*fp++,
	    p2=*fp++,
	    Dp2=*fp++;

	dpolystack.pop(2);


	if(instr->op=='^') {	    
	 		/* (p^a)' = a * p^(a-1) p' */
		Dp = pow(p1, p2 - 1.0) * p2*Dp1;
	    
	    p = pow(p1,p2);
	}else {
	    p=p1*p2;
	    
	
	    /* (p1*p2)' = p1'*p2 + p2'*p1 */
	    Dp = Dp1*p2+Dp2*p1;
	}
    }
   dpolystack.last(1)=p;
	dpolystack.last()=Dp;
        break;

    case 'n':
	dpolystack.last(1) *= -1;

        /* (-p)' = - p' */
	dpolystack.last() *= -1;

        break;
    default:
        assert(false);
    }
}

/*
 * calculate normal. calculate the gradient for each component
 */
flt
normal_evaluator::calc_all(svec<poly_instruction> const & instr)
{
    dpolystack.set_size(instr.sz());
    dpolystack.clear();
	
    /* calcalute the gradient. */
    for (int     i = 0; i < instr.sz(); i++) {
        exec(&instr[i]);
    }
    assert(dpolystack.sz() == 2);

    return dpolystack[1];
}


/* print out the stack for calculating the derivative */
void
normal_evaluator::print()
{
    #ifdef DEBUG
    dprintf("\n----\n");
    for (int     i = 0; i < dpolystack.sz(); i += 2)
    {
        dprintf("%d : %f, %f\n", i / 2, dpolystack[i], dpolystack[i + 1]);
    }
    #endif
}

/****************************************************************/


bool
algebraic_data::inside(vector x) const
{
    /* evaluate the polynomial in X */
    poly_evaluator pev( x,0);
    Poly tmp;
    pev(tmp, code);
    tmp.clean();
    assert(tmp.ord == 0);
    return (tmp.coef[0] > 0);
}

void
algebraic_data::print() const
{
#ifdef DEBUG
    dprintf("code: \n[");
    for (int i = 0; i < code.sz(); i++)
        (code + i)->print();
    dprintf("]\n");
#endif          /* DEBUG */
}

/* just the intersections. */
int
algebraic_data::do_intersection(ray *r, flt * rhits, int flags) const
{
    poly_evaluator pev(r->pos, r->dir);
    Poly   tosolve;
    pev(tosolve, code);

    /* and solve it */
    bool chkall = flags & CHKALL;

    int n = (daddy->flags & O_STURM) ?
	tosolve.sturm_solve_rt(rhits, chkall) : tosolve.solve_rt(rhits, chkall);

    return n;
}

vector
algebraic_data::normal(vector loc) const
{
    normal_evaluator nev(loc);
    vector  n =nev(code);
    return n;
}

shape_stats *algebraic_data::stat = shape_stats::regist("algebraic");
algebraic_data::algebraic_data(svec<poly_instruction> sv)
{
    type = ALGEBRAIC;
    code = sv;

}



void
algebraic_data::precompute()
{
    stat->howmuch++;
    prims++;
}


object *
        get_new_algebraic_object(svec<poly_instruction> sv)
{
    object *o = new object;
    o->data = new algebraic_data(sv);
    o->data->setdaddy(o);
    o->flags = O_INFINITE | O_CLOSED;
    return o;
}

#ifdef DEBUG
/* print instruction */
void
poly_instruction::print() const
{

    switch(op) {

   case 'x':
    case 'y':
    case 'z':
	dprintf("c %c ", op);
	break;
    case 'f':
	/* push a constant on the stack */
        dprintf("c %f ", factor);
	break;
    default:
    
        dprintf("op %c ", op);
    }
}

#endif          /* DEBUG */

/* eof */
