/*
 * bsp, adapted from GGems III by HWN
 */

#include	<stdio.h>
#include 	<stdlib.h>

#include	"ray.hh"
#include	"proto.hh"
#include	"ibuf.hh"
#include	"object.hh"
#include	"nonprim.hh"
#include	"opti.hh"

#include 	"extern.hh"

/****************************************************************
  CLASSES
 ****************************************************************/

struct Binnode
{
    bounding_box bbox;		/* extent of node */
    Optilist *members;          /* list of enclosed primitives */
    Binnode *child[2];          /* pointers to children nodes, if any */
    int division_axis;

    void print() const;
    void short_print() const;
            Binnode();
            ~Binnode();
    void add(const object *o) { members->add( o ); }
    bool    inside(const object *)  const;
    int     subdivide(int depth, int &maxdepth, int MaxListLength);
};

/* is obj  associated with this node? */
inline bool
Binnode::inside(const object * op) const
{
    return bbox.inside(*op->bbox);
}

Binnode::~Binnode()
{
    delete members;
    delete child[0];
    delete child[1];
}
Binnode::Binnode()
{
    members = new Optilist;
    child[0] = child[1] = 0;
    division_axis = -1;
}

void
Binnode::print() const
{
#ifdef DEBUG
    dprintf("BSP node ");
    if (child[0] || child[1])
	dprintf("(div %c) ", division_axis +'x');
    bbox.print();
    dprintf("{\n");
    if (child[0])
        child[0]->print();
    if (child[1])
	child[1]->print();
    if (members)
        members->print();
    dprintf("}\n");
#endif          /* DEBUG */
}

void
Binnode::short_print() const
{
#ifdef DEBUG
    dprintf("BSP node (div %c) #%d {\n", division_axis +'x',    (members)?members->sz():-1);
    assert(division_axis < 3 && division_axis >= 0);
    if (child[0])
        child[0]->short_print();
    if (child[1])
	child[1]->short_print();
    dprintf("}\n");
#endif          /* DEBUG */
}

/****************************************************************/

class Bintree : public object_list_optimisation
{
    Binnode *root;              // root of the entire bin tree
    Optilist *infinites;
public:
    void print() const ;
            Bintree(Optilist const &l);
            ~Bintree();
    bool    all_intersections(ibuf * , ray * theray, int , bool &) const;
};


Bintree::~Bintree()
{
    delete  root;
    delete  infinites;
}

void
Bintree::print() const
{
    #ifdef DEBUG
    dprintf("infinites: ");
    infinites->print();
    dprintf("tree: ");
    root->print();
    #endif
}


/***************************************************************
 TRACING
 ****************************************************************/

/*

  DESCRIPTION - calcs the distance between origin and plane, measured
  along the input direction. direction is a unit vector and Determines
  which of the half space of the two children contains pos, return
  that child as near, the other as far.


  PARAMS IN
  current:node  currently working on pos
  r: the ray
  c: coordinate to generate this function for.
  
  PARAMS OUT
  d:	distance
  near, far: near/far child, measured along ray.

  ALGORITHM plain plane intersection.
*/

//remember that child[0]->max or child[1]->min is the subdivision plane

PRIVATE inline
void get_division_info(int c, flt &d, Binnode * &near, Binnode * &far, ray * r, Binnode * current)
{
    flt mid = current->child[0]->bbox.max[c]; // UGLY! too many indirections ->
    if (mid > r->pos[c]) {
	near = current->child[0];
	far = current->child[1];
    } else {
	far = current->child[0];
	near = current->child[1];
    }

    if (r->dir[c])
	d = (mid - r->pos[c]) /r->dir[c];
    else
	d = INFTY;
}

/*
 * Data structure for a simple stack. This is necessary for implementing
 * an efficient linear BSP tree walking.
 */
struct stackelem
{
    Binnode *node;
    flt     min,
            max;
};

class bspstack
{
    stackelem *stack;
    int     max,
            ptr;
public:
    void       push(Binnode * &ch, flt & m, flt & M)
    {
        stack[ptr].node = ch;
        stack[ptr].min = m;
        stack[ptr].max = M;
        ptr++;
    }
    void            pop(Binnode * &Node, flt & m, flt & M)
    {
        ptr--;
        Node = stack[ptr].node;
        m = stack[ptr].min;
        M = stack[ptr].max;
    }
    void start()
    {
        ptr = 0;
    }
    void size(int s)
    {
        if (s > max)
        {
            max = s;
            delete  stack;
            stack = new stackelem[max * 3];	// why *3
        }
    }

    bspstack()
    {
        stack = 0;
        max = 0;
        ptr = 0;
    }
};

static bspstack the_bspstack;

/* trace a normal ray */
bool
Bintree::all_intersections(ibuf * globalq, ray * theray, int flags, bool &isinside) const
{
    flags |= USEMAIL | NO_CHKBOUNDS;
    ibuf   wait_queue;
    bool    quick = !(flags & CHKALL);


    flt     bhits[2];
    bool hit = infinites->all_intersections(globalq, theray, flags, isinside);
    int     n = intersect_rawbox(root->bbox.min, root->bbox.max, theray, bhits);

    flt     
            min,
            max;
    if (!n)
    {
        return false;
    } else if (n == 1)
    {
        min = 0;
        max = bhits[0];
    } else
    {
        min = bhits[0];
        max = bhits[1];
    }


    /* after having checked everything, 0 is popped of the stack */
    the_bspstack.start();

    {
        Binnode *current = 0;
        flt     bla = 0;

        the_bspstack.push(current, bla, bla);
    }

    Binnode *current = root,
           *near_child,
           *far_child;



    while (current)
    {
        while (current->child[0])
        {                       // descend to smallest children
	    flt dist;
            get_division_info(current->division_axis,dist, near_child, far_child, theray, current);

            if (ISZERO(dist))
            {                   /* whoops: standing on division plane. Act intelligently */
                the_bspstack.push(far_child, dist, max);
                current = near_child;
            } else if ((dist > max) || (dist < 0))
            {
                current = near_child;
            } else if (dist < min)
            {
                current = far_child;
            } else
            {
                the_bspstack.push(far_child, dist, max);
                current = near_child;
                max = dist;
            }
        }

        bool    gotcha = current->members->all_intersections(&wait_queue, theray, flags, isinside);
        if (gotcha)
        {                       // only check if new intersections were found

            hit |= true;

            // ????
            if (0 && wait_queue.mint() < max && wait_queue.mint() > min && quick)
            {
                /* only exit if intersection point is in the current node */
                break;
            }
        }
        the_bspstack.pop(current, min, max);
    }

    globalq->add_to(flags, wait_queue.min());


    return hit;
}

/****************************************************************
  INITIALISATION
 ****************************************************************/

/*
 * Builds the BSP tree by subdividing along the center of x, y, or z
 * bounds one each time this function is called.
 */
int        Binnode::subdivide(int depth,   /* current tree depth */
                                   int &maxdepth,       /* max reached depth */
                                   int maxlistlength)   /* the specified max allowed list length */
{
  int next_axis = (division_axis + 1) %3;
  members->clean();
    if (maxdepth < depth)
        maxdepth = depth;

    if (members->sz() <=  maxlistlength)
	return 1;
	    
    flt     m = 0.5;    // adjust m to alter division behaviour
    for (int childcount = 0; childcount < 2; childcount++) {
            flt     minfactor = (m * childcount);
            flt     maxfactor = ((1 - m) * (1 - childcount) + childcount);

            Binnode *childp = new Binnode;
            child[childcount] = childp;

            childp->bbox = bbox;

	    /* current subdivision plane is x */
	    int a = division_axis;
	    childp->bbox.min[a] =
		bbox.min[a] + minfactor * (bbox.max[a] - bbox.min[a]);
	    childp->bbox.max[a] =
		bbox.min[a] + maxfactor * (bbox.max[a] - bbox.min[a]);
	    childp->division_axis = next_axis;
	    /* put stuff in children */
	    /* one object can be in *both* children */
	    for (int i = 0; i < members->sz(); i++) {
	      if (childp->inside( (*members)[i]) ){
		childp->add( (*members)[i] );
	      }
	    }
	  }


#if 0
    for (int i=0; i < 2; i++) {
	child[i]->members->update();
	child[i]->bbox.intersect(child[i]->members->bbox);
    }   
#endif

    int threshold = (members->sz() * 9) / 10;
    int leafs;

    if (child[0]->members->sz() > threshold && child[1]->members->sz() > threshold)
        {
            /* this step wasn't really useful.. */
            delete  child[0];
            delete  child[1];
            child[0] = 0;
            child[1] = 0;

            /* Try again. Beware. this is recursion maxlistlen is increased to guarantee return */
	    division_axis =  next_axis;
            leafs = subdivide(depth + 1, maxdepth, (maxlistlength * 5) / 4 + 1);
        } else {
		delete  members;    // everything was moved into children. No longer needed.
		members = 0;
		leafs = child[0]->subdivide(depth + 1, maxdepth, maxlistlength);
		leafs += child[1]->subdivide(depth + 1, maxdepth, maxlistlength);
	    }

    return leafs;
}

Bintree::Bintree(Optilist const &l)
{
    root = new Binnode;
    infinites = new Optilist;

    for (int i = 0; i < l.sz(); i++)
	{
	    if (l[i]->bbox) {
		if (l[i]->bbox->empty()) 
		    warning("will ignore empty boundingboxes");
		else 
		    root->members->add(l[i]);
	    }
	    else 		
		infinites->add(l[i]);
	}

    infinites->clean();
    root->bbox = root->members->bbox;
    root->division_axis = X;
    int     max = 0;

    /* Start building the BSPTree by subdividing along the x axis first */
    int     leafs = root->subdivide(0, max, the_scene.bunching);

    printf("Calculated BSP. Depth %d, leafs %d\n", max, leafs);
    max += 2;
    the_bspstack.size(max);
}

/*
 * Initialize and start the building of BSPTree.
 */
object_list_optimisation *
init_BSP(Optilist const &l)
{
    return new Bintree(l);
}

