/*
   camera.cc -- standard routines for camera

   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */


#include "proto.hh"
#include "extern.hh"
#include "camera.hh"
#include "grid.hh"
#include "ibuf.hh"



/// generate an eyeray
void
Camera::set_eyeray(ray & r, flt & filter, int curx, int cury, int xres, int yres)
{
    // computation is done in a 2 x 2 rectangle around the origin.
    flt     pix_width = 2.0 / flt(xres);
    flt     pix_height = 2.0 / flt(yres);

    // in pixelfiles screen starts upper left, in stead of bottom left
    cury = yres - cury;	
    

    // xr and yr are in [-1,1] range. r stands for relative. (-1,-1) is upper left corner
    flt     xr = (curx+0.5) * pix_width - 1.0;
    flt     yr = (cury+0.5) * pix_height - 1.0;

    // sample xa and ya
    flt     xa,
            ya;
    the_scene.the_grid->sample(xa, ya, filter, r.type);
    xa *= pix_width / 2.0;
    ya *= pix_height / 2.0;

    /* ray direction calculations */
    vector  hor = (xr + xa) * fov * aspect* right;
    vector  ver = (yr + ya) * fov * up;
    r.pos = eye;                /* eye is the beginning of the ray */
    r.dir = eye_dir + hor + ver;
    r.dir /= r.dir.length();
}

/**
  (default camera : +X is right, +Y is up, +Z is into screen)
  */
void
Camera::set_fov_angle(flt angle)
{
    fov = eye_dir.length() * tan(degtorad(angle) / 2);
}

/* point the camera towards a particular point. */
void
Camera::point(vector lookat)
{
    flt     dirlen;
    dirlen = (eye_dir).length();
    eye_dir = lookat - eye;
    eye_dir /= eye_dir.length();
    eye_dir = dirlen * eye_dir;
}


// defaults.
Camera::Camera()
{
    eye_dir.set(0, 0, 1);
    eye.set(0, 0, -10);
    sky.set(0, 1, 0);
    set_fov_angle(90);
    aspect = 1.0;
}

void
Camera::translate(const vector t)
{
    eye += t;
}

void
Camera::precompute()
{
    /* init the camera */
    right = sky^eye_dir;      /* the x screen vector */

    if (!right)
        errormsg("direction vector can't be sky vector.");
    right /= right.length();

    up = eye_dir ^ right;       /* the y screen vector */
    up /= up.length();
}

void
Camera::rotate(const matrix & rotmat)
{
    eye_dir.rotate(rotmat);
    eye.rotate(rotmat);
    sky.rotate(rotmat);
}

void
Camera::scale(vector s)
{
    eye_dir.scale(s);
    eye.scale(s);
    sky.scale(s);
}

void
Camera::print() const
{
#ifdef DEBUG
    dprintf("camera {\n");
    eye_dir.print("eye_dir");
    eye.print("eye");
    sky.print("sky");
    dprintf("fov %f\n", fov);
    dprintf("aspect %f\n}\n", aspect);
#endif          /* DEBUG */
}
