/*
   cone.cc -- handle cone type objects

   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */


#include "object.hh"
#include "shapes.hh"
#include "proto.hh"
#include "extern.hh"
#include "ibuf.hh"
#include "poly.hh"


/* initialize a cone_data */


shape_stats *cone_data::stat = shape_stats::regist("cone");

cone_data::cone_data(vector pp1, vector pp2, flt R1, flt R2)
{
    type = CONE;
    p1=pp1;
    p2=pp2;
    r1 = R1;
    r2 = R2;
    locap = hicap = true;
}
/* is vector loc inside a cone ? */
bool
cone_data::inside(vector loc) const
{
    loc -= p1;
    flt         d = axis * loc;
    if (d < 0 || d > axlen)
        return false;
    vector      v = loc - (d * axis);

    flt    R = abs(d * slope + r1);
    return (v.length() < R);
}
/* scale it. */
void
cone_data::scale(const vector s)
{
    flt             factor = abs(s.x);
        r1 *= factor;
        r2 *= factor;
        p2.scale(s);
        p1.scale(s);
}

/* rotation */
void
cone_data::rotate(const matrix & rotmat)
{
        p1.rotate(rotmat);
        p2.rotate(rotmat);
}
/* and translation */
void
cone_data::translate(const vector t)
{
        p1 += t;
        p2 += t;
}
void
cone_data::print() const
{
#ifdef DEBUG
    p1.print("p1");
    p2.print("p2");
    dprintf("r1 %f, r2 %f, cap: lo %d, hi %d.", r1, r2, locap, hicap);
#endif          /* DEBUG */
}
/*
 * the real intersection calculation.
 */
int
cone_data::do_intersection(ray * r, flt * rhits, int) const
{
    vector   pos = r->pos - p1;       /* translate to origin */

    flt dotdir = r->dir * axis;
    vector    projdir = -dotdir * axis;
    projdir += r->dir;

    flt    dotpos = pos * axis;
    vector  projpos = -dotpos * axis;
    projpos += pos;


    flt lin1 = slope * dotdir;
    flt	cons1 = slope * dotpos + r1;
    Poly   p;
    p.ord = 2;
    p.coef[2] = (projdir * projdir);
    p.coef[1] = 2 * (projdir * projpos);
    p.coef[0] = (projpos * projpos);
    p.coef[2] -= sqr(lin1);
    p.coef[1] -= 2 * lin1 * cons1;
    p.coef[0] -= sqr(cons1);

        if (ISZERO(p.coef[2]))
        {
            return 0;
        }
    flt                infroots[2];
    int    infhits = solve_rt_quadric(p.coef, infroots);

    int    totalhits = 0;
    for (int i = 0; i < infhits; i++)
    {

        flt d = dotdir * infroots[i] + dotpos;
        if (d > 0 && d < axlen)
        {
            rhits[totalhits++] = infroots[i];
        }
    }

    if (totalhits == 2)
        return 2;

    /* try caps */
    if (locap && !ISZERO(dotdir))
    {
        flt t = -dotpos / dotdir;
        if (t > tolerance)
        {
            vector g = t * r->dir;
            g += pos;
            if (r1sq > g * g)
                rhits[totalhits++] = t;
        }
    }
    if (hicap && !ISZERO(dotdir))
    {
        flt t = (axlen - dotpos) / dotdir;
        if (t > tolerance)
        {
            vector g = t * r->dir;
            g += pos;
            vector h = -axlen * axis;
            g += h;
            if (r2sq >  g * g)
                rhits[totalhits++] = t;
        }
    }
    assert(totalhits <= 2);
    if (!totalhits)
        return 0;
    if (totalhits == 2)
    {
        if (rhits[0] > rhits[1])
        {
            flt t = rhits[0];
            rhits[0] = rhits[1];
            rhits[1] = t;
        }
        if (rhits[1] < tolerance)
            return 0;
        else if (rhits[0] < tolerance)
        {
            rhits[0] = rhits[1];
            return 1;
        } else
            return 2;

    } else if (totalhits == 1)
        return (rhits[0] >= tolerance);

    return false; // unreachable
}


vector
cone_data::normal(vector loc) const
{
    vector  n;

    loc -= p1;
    flt d = loc * axis;

    if (ISZERO(d))
        n = -axis;
    else
    {
        if (ISZERO(d-axlen))
            n = axis;
        else
        {
            vector v = d * axis;
            v = loc - v;
            v /= v.length();
            vector w = -slope * axis;
            n = v + w;
        }
    }
    return n;
}
#include "matrix.hh"

void
cone_data::bound(matrix   *trans, bounding_box *bbox)
{
    /* do autobounding, End caps determine bounding box */
    vector rtv(M_PI / 2, 0, 0);
    matrix ctm;
    ctm.rotate(rtv);

    flt         Th = acos(axis.z / (axis).length());
    flt    Ph = atan2(axis.y, axis.x);
    vector v(0, Th, Ph);
    matrix rm;
    rm.rotate(v);

    rm = rm * ctm;

    matrix tm;
    tm.translate(p1);
    ctm = tm * rm;

    if (trans)
	ctm = (*trans) * (ctm);
    bound_torus(bbox, r1, 0, ctm);

    tm.translate(p2);
    ctm = tm * rm;
    if (trans)
	ctm = (*trans) * (ctm);

    bounding_box box2;
    bound_torus(&box2, r2, 0, ctm);

    bbox->add(box2);
}

void
cone_data::precompute()
{
    stat->howmuch++;

    /* do precomputation */
    if (r1 < r2)
    {
        flt     r = r1;

        r1 = r2;
        r2 = r;

        vector	p = p1;
        p1 = p2;
        p2 = p;
    }
    r1sq = sqr(r1);
    r2sq = sqr(r2);
    axis = p2 - p1;
    axlen = (axis).length();
    slope = (r2 - r1) / axlen;

    axis /= axis.length();

    /* stats */
    prims++;
}


/* alloc/init */
PUBLIC object *
        get_new_cone_object(vector v1,vector v2,flt r1,flt r2)
{
    object *o = new object;
    o->data = new cone_data(v1,v2,r1,r2);
    o->data->setdaddy(o);
    o->flags = O_CLOSED | O_FINITE_INSIDE | O_ONLY_UNIFORM_SCALE ;

    return o;
}
/* eof */
