/*
  gif.cc -- read a gif file. scooped from the Linux gif viewer spic by
  Harm Hanemaayer <hhanemaa@cs.ruu.nl>

  Thoroughly rehacked, 18 sep '95
  objectified, elegantised, readabilified and (hopefully) robuster by hwn
  helpful hints from code by Steven A Bennet.
 */

#include <string.h>
#include "pixel.hh"
#include "proto.hh"
#include "texture.hh"

/****************************************************************/
typedef unsigned long ulong;
const ulong    clear_code(int r) { return (1 << r); };
const ulong    end_of_input(int r) { return clear_code(r) + 1; }


struct table_ent
{
    ulong prefix;
    ulong k;
};

const int GTABSIZE = 4096;

class code_table_struct {
    table_ent tab[GTABSIZE];
    int rootsize;
    
 public:
    code_table_struct(byte);
    int len;

    void reset();
    void get_string(ulong incode, byte * s, ulong &len) const;
    void addstring(ulong prefix, byte ink);
};

code_table_struct::code_table_struct(byte rs)
{
    rootsize = rs;
    int CC = clear_code(rs);
    int EOI = end_of_input(rs);

    for (int i = 0; i < CC; i++) {
        tab[i].prefix =GTABSIZE;		// past the end of table
        tab[i].k = i;
    }

    // clear_code and end_of_input are  reserved
    for (int  i= EOI + 1; i < GTABSIZE; i++)
    {
        tab[i].prefix = GTABSIZE;
        tab[i].k = 0;
    }
    
    reset();

}


/// reset to  initial condition
void
code_table_struct::reset()
{
    len = end_of_input(rootsize) + 1;
}

/// fetch string from table. Result in s, length(s) in len
void
code_table_struct::get_string(ulong incode, byte * s, ulong &slen) const
{
    assert(incode  < len);
    byte    stringbuf[GTABSIZE];
    int i = 0;
    int j = incode;
    while (j < GTABSIZE)
    {
        stringbuf[i] = tab[j].k;
        j = tab[j].prefix; // keep following indirections until in {0,..,clearcod - 1}
	assert(0 <= j && j <= GTABSIZE);
        i++;
    }

    slen = i;
    assert(i>0 && slen < GTABSIZE);
    // reverse string
    while (i--)
    {
        *s++ = stringbuf[i];
    }
    
}

/// add prefix, suffix to table.
void
code_table_struct::addstring(ulong prefix, byte suffix)
{
    tab[len].prefix = prefix;
    tab[len].k = suffix;
    len++;
}

/*****************INPUT ROUTINES*****************/

/// image data is raw byte(s)-per-pixel
static void
output_string(byte *&outputp, byte * s, ulong len)
{
    memcpy(outputp, s, len);
    outputp += len;
}

class bitbuffer {    
    byte data[512];
    ulong len,bitindex,mask;    
    FILE *infile;
    byte codesize,rootsize;
    
    unsigned int byteindex() { return (bitindex >> 3); }
    unsigned int shift() { return (bitindex & 7); } // bit count inside byteindex

    // read a GIF block: 1 byte blocksize, then block
    void bitread() {	
	byte count = fgetc(infile);

	byte *bp;
	
	if (len > 0) {// >1?
	    // copy remember first 2 bytes
	    bitindex -= (len - 2) * 8;
	    assert(16 > bitindex && bitindex>0);
	    
	    data[0] = data[len -2];
	    data[1] = data[len -1];
	    len = count + 2;
	    bp = data+2;
	} else {
	    len = count;
	    bp = data;
	}
	while (count--)
	    *bp++ = fgetc(infile);
    }

public:
    ulong get_next_code () {

	// haven't got codesize # bits left. Get more bits
	if (bitindex + codesize >= len * 8)
	    bitread();

	// destill the next codesizebits in buffer.
	// slow but endian independent
	ulong twobytes = ( int(*(data+byteindex())) << 8)  | *(data+byteindex()+1); 
	ulong bpb =  twobytes >> shift();
	ulong code = (ulong) (bpb & mask);
	bitindex += codesize;
	printf("codsiz %d\n", codesize);
	printf("idx %lu, cod %lu\n", bitindex, code);
	return code;
    }
    void try_grow(int tablelen) {

	// codetable.len == 2^codesize.
	// It should get bigger for accomodating more codes.

	if (tablelen == (1 << codesize)) {
	    if (codesize == 12)
		errormsg("help: too many codes.");
	    codesize++;	    
	    mask = (1 << codesize) - 1;
	}
    }

    bitbuffer(FILE *f,byte r) {
	rootsize =r;
	infile = f;
	bitindex = 0;
	len = 0;
	reset();
    }
    void reset() {	
	codesize = rootsize + 1;
	mask = (1 << codesize) - 1;	// mask to filter CODESIZE least significant bits
    }
};

static void
create_picture(byte *outputp , FILE * f)
{
    byte rootsize = fgetc(f);

    if (rootsize < 2 || rootsize > 9)
	errormsg("bad code size");

    const ulong CC = clear_code(rootsize);
    const ulong EOI = end_of_input(rootsize);    
    code_table_struct codetable(rootsize);

    bitbuffer buf(f, rootsize);
    
    // Skip first code . should be clear_code
    //    ulong bitindex = 16 + codetable.codesize;   

    ulong code;
    ulong old = CC;
    while ((code = buf.get_next_code()) != EOI) 
    {
	if (code == CC) {
	    printf("clear\n");
	    codetable.reset();
	    buf.reset();
	    continue;
	} else if (old == CC) {
	    assert(code < codetable.len && code != CC);
	    output_string(outputp, (byte *) (&code), 1); // portable?
	    old = code;
	    continue;
	}
	byte    s[GTABSIZE];
	ulong len;
	if (code < codetable.len) {			/* Code exists in table */
	
	    codetable.get_string(code, s, len);
	    output_string(outputp, s, len);

	    // suffix s[0] is  the oldest ("primitive") code	    
	    codetable.addstring(old, s[0]);	    
	} else if (code == codetable.len) {		/* Code does not yet exist */
	
	    codetable.get_string(old, s, len);
	    codetable.addstring(old, s[0]);	    
	    s[len++] = s[0]; 	//?

	    output_string(outputp, s, len);
	} else {
	    // something more intelligent
	    errormsg("create_gif bad code");
	}
	buf.try_grow(codetable.len);
	old = code;
    }
}

/* load a gif into a pixmap data structure */
static void
load_gif(colmapped_data * image, FILE * f)
{
    /* read header */
    byte    inbuf[256];

    // gif signature
    fread(inbuf, 6, 1, f);
    if (memcmp(inbuf, "GIF", 3))
        errormsg("File not in GIF format.\n");

    // header: read logical screen descriptor (LSD) 
    fread(inbuf, 7, 1, f);

    // global colormap
    //byte c = (byte) inbuf[4];
    long        gct = inbuf[4] & 128;      /* Global color table present ? */
    long       gct_size = inbuf[4] & 7;   /* size of GCT */

    /* read image descriptor */
    if (gct)
    {                           /* skip GCT */
        int j = 2;
        for (int i = 0; i < gct_size; i++)
            j *= 2;

        /* Palette data is 256 red, green and blue values (0-63) */
        image->pal = new byte_color[j];
        image->palsize = j;

	// RGB
        fread((byte*)image->pal, 3 * j, 1, f);
    } else
        image->palsize = 0;     /* is this ok? */

    fread(inbuf, 10, 1, f);

    // ignoring image separator ','
    // ignoring imageleft,imagetop (4 bytes)
    // size (lsb first)
    image->imagex = int(inbuf[5]) | (int( inbuf[6]) << 8);
    image->imagey = int(inbuf[7]) | (int( inbuf[8]) << 8);


    long    lct = inbuf[9] & 128;      // use local color table?
    long    lct_size = inbuf[9] & 7;   // lct_size bits per pixel.
    if (lct)
    {                           /* Skip LCT */
	assert (!gct);
        int j = 2;

	for (int i = 0; i < lct_size; i++)
	    j *= 2;

        image->pal = new byte_color[j];
        image->palsize = j;

	// RGB
        fread((byte*)image->pal, 3 * j, 1, f);
    }

    
    int s = image->imagex * image->imagey;
    image->data = new byte[s];
    create_picture((byte*) image->data, f);
    // ignoring trailing info

}


/* read in a GIF file. */
PUBLIC pixel_data *
        get_GIF_pixels(char *fn)
{
    pixel_data *q =search_pixels(fn);
    if (q) 
	return q;
    
    colmapped_data *p =  new colmapped_data(fn);
    add_pixels(p);
        
    FILE   *f = open_file(fn, "rb");
    char   *buf = new char[BUFFERSIZE];
    if (setvbuf(f, buf, _IOFBF, BUFFERSIZE))
        errormsg("cannot buffer TGA input");

    load_gif(p, f);

    close_file(f);
    delete[] buf;
    printf("]");    
    return p;
}
/* eof */
