/*
 * imagemap.cc -- read imagemaps.

   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */

#include <string.h>

#include "object.hh"
#include "texture.hh"
#include "proto.hh"
#include "ibuf.hh"
#include "pixel.hh"
#include "matrix.hh"
/***************************************************************************
 * INVERSE MAPS								   *
 ***************************************************************************/

/*
 * give inverse mapping for point, on sphere centered in 0,0,0
 *
 * z = cos(v * pi), x = sin(v*pi)*cos (u*2pi), y = sin(v*pi)*sin(u*2pi).
 */
static void
        inverse_spheremap(flt & u, flt & v, vector x)
{
    x.normalise();
    u = acos(x.z) / M_PI;
    v = safe_arctangent(x.y, x.x) / (2 * M_PI) + 0.5;
}
/*
 * maps x onto plane.
 */
static void
        inverse_planemap(flt & u, flt & v, vector x)
{
    u = x.x;
    v = x.y;
}
/*
 * maps x onto torus in XZ plane, major radius 1.0
 */
static void
        inverse_torusmap(flt & u, flt & v, vector x)
{
    
    u = safe_arctangent(x.z, x.x) / (2 * M_PI) + 0.5;
    flt     horiz = 1.0 - sqrt(sqr(x.x) + sqr(x.z));
    v = safe_arctangent(x.y, horiz) / (2 * M_PI) + 0.5;
}
/*
 * maps x onto a cylinder with axis along the Z-axis.
 */
static void
        inverse_cylindermap(flt & u, flt & v, vector x)
{
    u = safe_arctangent(x.y, x.x) / (2 * M_PI) + 0.5;
    v = x.z;
}
/****************************************************************/
void
image_map::set_type(int j)
{
    type  = maptype(j);
    switch (j)
    {
    case 0:
        inv_map = inverse_planemap;
        break;
    case 1:
        inv_map = inverse_spheremap;
        break;
    case 2:
        inv_map = inverse_cylindermap;
        break;
    case 5:
        inv_map = inverse_torusmap;
        break;
    default:
        errormsg("unknown maptype");
    }
}
/*
   OTHER STUFF
 */
image_map::image_map(pixel_data *p)
{
    inv_map = inverse_planemap;
    pic = p;
    u_range = v_range = 1.0;
    u_offs = v_offs = 0.0;
    uvswap = once = interpolated = false;
    usenormal = false;
}


image_map::~image_map()
{
    pic->refs--;
    assert(pic->refs >= 0);
    if (pic->refs == 0) {
        delete  pic;
        pic = 0;
    }
}

/* lookup the color in an image map. */
color
image_map::get_color(vector x, vector n) const
{
    flt     u,
            v;

    inv_map(u, v, (usenormal) ? n : x);

    /* swap them? */
    if (uvswap)
    {
	flt w = u;
        u = v;
        v = w;
    }
    /* transformations */
    u *= u_range;
    v *= v_range;
    u -= u_offs;
    v -= v_offs;

    if (once && (u < 0.0 || u > 1.0 || v < 0.0 || v > 1.0)) {
        return color(0);
    }
    /* truncate, only the fractional part remains */
    u -= ceil(u);
    if (u < 0)
        u += 1.0;

    v -= ceil(v);
    if (v < 0)
        v += 1.0;

    v = 1.0 - v;                /* 0,0 is left upper with pictures. */

    return color ((interpolated) ? pic->interpolate( u, v) :
		  pic->get_color(u,v));
}



void
image_map::_print() const
{
#ifdef DEBUG
    pic->print();
    dprintf("swap %d usenormal %d maptype %d\n", uvswap, usenormal, type);
    dprintf("uvrange %f %f offset %f %f\n", u_range, v_range, u_offs, v_offs);
#endif // DEBUG
}
/* eof */
