/*
   lights.cc --  routines for lights


   (c) 1996 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */



#include "object.hh"
#include "proto.hh"
#include "extern.hh"
#include "ibuf.hh"
#include "color.hh"
#include "stats.hh"
#include "lightbuf.hh"



/// alloc pixels
lightbuf_plane::lightbuf_plane() {
    pixels = new lightbuf_pixel*[lightbufres];
    for (int i =0  ; i < lightbufres; i++)
	pixels[i] = new lightbuf_pixel[lightbufres];
}

lightbuf_plane::~lightbuf_plane () {
    for (int i =0  ; i < lightbufres; i++)
	delete[] pixels[i];
    delete [] pixels;
}

void
lightbuf_plane::print() const
{
#ifdef DEBUG
    for (int i =0  ; i < lightbufres; i++)
	for (int j=0  ; j < lightbufres; j++)
	    pixels[i][j].print();
#endif
}

/// no constructor due to array.
void
lightbuf_plane::init(vector normal) {
    axis = normal.dominant();
}
/**
   normal: outward normal to plane
*/


///  lookup pixel in square array
 
lightbuf_pixel *
lightbuf_plane::find_pixel(vector dir) const
{
    assert( dir.dominant()  == axis );
    flt x = dir[(axis + 1) % 3 ];
    flt y = dir[(axis + 2) % 3 ];
    assert(abs(x) <= 1 && abs(y)<=1);

    // map onto [0,1]
    x = x/2+0.5;
    y = y/2+0.5;
    
    int ix = (int) floor(x*lightbufres);
    int iy = (int) floor(y*lightbufres);

    if (ix == lightbufres)
	ix--;
    if (iy ==lightbufres)
	iy--;
    return &(pixels[ix][iy]);
}

/** IN: direction d, relative to center
 */

/****************************************************************/
void
lightbuf::print() const
{
#ifdef DEBUG
    dprintf("lbuf {\n");
    for (int i = 0; i<6; i++)
	faces[i].print();
    dprintf("lbuf }\n");
#endif
}

lightbuf::lightbuf(vector c)
{
    center = c;
    
    for (int i=0; i <3; i++) {
	vector v(0);

	v[i] = -1.0;
	faces[i].init(v);
	v[i] = 1.0;
	faces[i+3].init(v);
    }
}

// look up the pixel in this lighbtuf.
lightbuf_pixel *
lightbuf::find_pixel(ray *r) const
{
    axes a = r->dir.dominant();
    int j =a;
    flt dominant_component = r->dir[a];
    if (dominant_component > 0) {
	j+=3;
    } 
    
    vector x = r->dir / dominant_component;
              
    return (faces + j)->find_pixel(x);
}

bool
lightbuf::test(ray *r) const
{
    return find_pixel(r)->test(r);
}

/****************************************************************/
/*
 * Shadow check.  Check if the ray from the intersection point to the
 * light, intersection any objects inbetween.  If it does, we have a
 * shadow. One improvement that can be made is to check the transparencies
 * of all the intersecting objects and make the shadow have a different
 * intensity depending on the transparency of the objects
 */
static shape_stats *cache = shape_stats::regist("shadow cache",-1);

bool
lightbuf_pixel::test(ray *r) const
{
    ibuf q;					  // for intersect routine
    bool b;					  // dummy for intersect routine

    // shadow buffer
    cache->test++;
    for (int i=0; i < shadowers.sz(); i++)	
	if (shadowers[i]->all_intersections(&q, r, USEMAIL, b) && q.mint() < r->mint) {
	    cache->hit++;
	    return true;
	}

    /* check whole scene */
    if (the_scene.the_object->all_intersections(&q, r, USEMAIL, b))
    {
	object *p = q.min().obj;
	flt t = q.mint();
	
        /* found intersection */	
        r->line.rayintersections++;

        if (t < r->mint) {	/* something is blocking the light */
	    object *dp = p;	
    
	    assert(p->data->type != COMPOSITE);
	    while (dp->data->type != COMPOSITE) { p= dp; dp = dp->daddy; }
					   // don't want to break in middle of CSG

	    /* add to cache if not found yet. */
	    int j = 0;
	    for( ; j < shadowers.sz(); j++)		
		if (shadowers[j] == p )
		    break;
	    if (j == shadowers.sz() )
		((lightbuf_pixel *)this)->shadowers.add(p);
	    
	    return true;
        }
    }
    return false;    
}

void
lightbuf_pixel::print() const
{
#ifdef DEBUG
    for (int i=0; i < shadowers.sz(); i++)	
	shadowers[i]->short_print();
#endif
}
