// -*-c++-*-
#ifndef OBJECT_H
#define OBJECT_H
#include "vector.hh"

struct transpair
{
    /// inverse transformation (if applicable) 
    matrix *inv_trans;
    /// transformation; killed after precomputation. 
    matrix *trans; 


    transpair();
    transpair(transpair & src);
    void precompute();
    ~transpair();
    void print() const ;

    void rotate(const matrix & rotmat);
    void translate(const vector t);
    void scale(const vector s);
private:
    void allocself();
};


/// for the shape->type variable 
enum shapetype
{
    NOSHAPE, SPHERE, QUADRIC, PLANE, LIGHT, BOX,
    TORUS, CSGUNION, CSGINTER, COMPOSITE, ALGEBRAIC, SUPERQ,
    POLYGON, EXTRUSION, TRIANGLE, DISC, CONIC, CONE
};

struct bounding_box {
    vector min,max;

    bounding_box();
    bounding_box(const vector &, const vector &);

    void  set_empty();
    void  set_full();
    bool empty() const;
    void add(vector);
    void add(const bounding_box);
    bool inside(bounding_box const &) const;
    bool inside(vector const &) const;
    void intersect(const bounding_box  &);
    void print() const ;
    void scale(const flt fudge);
    void add_speed(const vector);
};

/****************************************************************
  SHAPE
 ****************************************************************/
struct shape_data
{
    //*************** DATA
    shapetype type;
    int refcount;
    object *daddy;
    //*************** STATS
    static long prims,
            nonprims;
    // ************** METHODS for derived classes
    virtual shape_stats *getstatp() const = 0 ;
    shape_data();
    virtual void bound(matrix *, bounding_box *);
    virtual void add_children_to_optilist(Optilist *);
    /// virtual copy constructor
    virtual shape_data *clone() = 0;	
    virtual bool all_intersections(ibuf * q, ray * r, int flags, bool &b) const ;
    virtual int do_intersection(ray * r, flt *, int flags) const ;
    virtual void precompute() = 0;
    virtual ~shape_data();
    virtual vector normal(vector) const;
    virtual bool inside(vector) const;
    virtual void rotate(const matrix & rotmat);
    virtual void translate(const vector t);
    virtual void scale(const vector s);
    virtual void print() const ;
    virtual void add_child(object *);
    void setdaddy(object const * dad);
    virtual void add_speed(vector);

};

/****************************************************************/

class object_list {
    friend class object_iter;
    object *first;
public:

    operator bool() { return first; }
    void append_to(object * toadd);
    void operator = (object_list &bla);
    object_list(object_list &);
    object_list();
    ~object_list();
    void rotate(const matrix & rotmat);
    void precompute(object *);
    void translate(const vector t);
    void scale(const vector s);
    void bound();
    void add_speed(const vector s);
    void print() const;
    bool empty() const { return !first; }
    void add_to(object *o);
    void append_to(object_list *o);
};


/****************************************************************
 * the OBJECT structure.					*
 ****************************************************************/

///the OBJECT structure. 
class object
{
friend class object_iter;
friend class object_list;
    /// for link list 
    object  *next;
    /// You have new mail in /usr/object/last_ray_id :) */
    long    last_ray_id;
    transpair tranp;
    /// motion coefficients */
    vector *speed;  


    bool intersect_bounds(ibuf *globalq, ray *r, int traceflags, bool &isinside) const;
    bool intersect_clips(ibuf *globalq, ray *localray, ray *r, int traceflags, bool &isinside) const;

public:
    static shape_stats *stat,
           *boundstat,
           *clipstat;
    
    //********************************* DATA

    shape_data *data;

    /// for bounding 
    bounding_box *bbox;  

    /// for texturing 
    object *daddy;
    /// bounding shapes to the object 
    object_list bounds;
    /// clipping shapes 
    object_list           clip;  

    /// miscellaneous info bits 
    int     flags;
    /// the texture of the object 
    texture *text;  

    //********************************** METHODS
    vector getspeed();
    vector transformd(const vector & d) const;
    vector transformx(const vector & x) const;
    vector transformn(const vector & n) const;
    void transformr(ray & r) const;
    object();
    ~object();
    object(object &);
    bool    all_intersections(ibuf * globalq, ray * r, int flags, bool & isinside) const;

    void add_speed(const vector s);

    void add_bounds_to_clip();
    void precompute();
    void add_to_optilist(Optilist *l);

    bool inside(vector loc) const;
    void rotate(const matrix & rotmat);
    void translate(const vector t);
    void scale(const vector s);
    void bound();

    void print() const ;
    void short_print() const;
    vector normal( vector loc) const;
    void append_to(object *o);


    //************************************ STATS
};

/// walk through a list of objects
class object_iter {
    object *p;
public:
    bool ok() { return p;}
    object_iter(const object_list &l) { p= l.first ; }
    object *operator->()	{ return p; }
    object &operator*()	{ return *p; }
    operator object *() { return p; }
    operator bool() { return bool(p); }
    object *next() {
	if (p) return p = p->next;
	else return p;
    }
};
#endif
