/*
   optimise.c -- optimisation utility routines

   (c) 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */


#include "object.hh"
#include "extern.hh"
#include "shapes.hh"
#include "ibuf.hh"
#include "proto.hh"
#include "opti.hh"


void
Optilist::add(object const * op)
{
    assert(op);

    // add pointer
    svec<object const *>::add( op);
    // update bbox.
    
    if (op->bbox) // added obj. is infinite
	bbox.add(*op->bbox);
    else
	bbox.set_full();
}
/* recalculate extent of list */
void
Optilist::update()
{
    bbox.set_empty();
    for (int i = 0; i < sz(); i++)
	bbox.add(* (*this) [i]->bbox);
}

Optilist::Optilist()
{
    bbox.set_empty();
}

/* build a list from o, and everything in it */
Optilist::Optilist(shape_data *c)
{
    bbox.set_empty();
    c->add_children_to_optilist(this);
}

/* reduce memory used by this one. */
void
Optilist::clean()
{
    svec<object const *>::precompute();
}

/*
 * This routine intersects ray with all of the objects in the optilist and
 * returns the closest intersection distance and the interesting object,
 * if there is one.
 */
bool
Optilist::all_intersections(ibuf * q, ray * r, int flags, bool &b) const
{
    bool    gotcha = false;
    /* walk the element list, and intersect each element */
    for (int i = 0; i < sz(); i++)
    {
        if ((*this)[i]->all_intersections(q, r, flags | USEMAIL, b))
            gotcha = true;

        if (q->mint() < r->mint)/* early exit possible? */
            break;
    }
    return gotcha;
}

/* and a debugging tool */
void
Optilist::print() const
{
#ifdef DEBUG
    int     i;
    dprintf("len %d\n", sz());
    for (i = 0; i < sz(); i++)
        (*this)[i]->short_print();
#endif          /* DEBUG */
}
/********************BBOX***********************/

bounding_box::bounding_box() {
   set_full();
}

bounding_box::bounding_box(const vector &v1, const vector &v2) {
    min = v1;
    max = v2;
}

void
bounding_box::set_empty() {
    min.set(INFTY);
    max = -min;
}

void
bounding_box::set_full() {
    max.set(INFTY);
    min = -max;
}

void
bounding_box::print() const {
    #ifdef DEBUG
    min.print("min");
    max.print("max");
    #endif
}

void
bounding_box::add_speed(const vector s) {
    vector m = min, M=max;

    set_empty();
    add(m + the_scene.start*s);
    add(M + the_scene.start *s);
    add(m + the_scene.stop*s);
    add(M + the_scene.stop*s);
}

void
bounding_box::scale(const flt s) {
    vector diag = (max - min)/2 + vector(EPSILON);
    vector center  = (max + min)/2;
    max = center + s*diag;
    min = center - s*diag;
}

void
bounding_box::intersect(const bounding_box &b)
{
    min.x = MAX(min.x, b.min.x);
    min.y = MAX(min.y, b.min.y);
    min.z = MAX(min.z, b.min.z);
    max.x = MIN(max.x, b.max.x);
    max.y = MIN(max.y, b.max.y);
    max.z = MIN(max.z, b.max.z);
}

void
bounding_box::add (bounding_box bbox)
{
    add(bbox.min);
    add(bbox.max);
}


void
bounding_box::add(vector point)
{
    min.x = MIN(min.x, point.x);
    min.y = MIN(min.y, point.y);
    min.z = MIN(min.z, point.z);
    max.x = MAX(max.x, point.x);
    max.y = MAX(max.y, point.y);
    max.z = MAX(max.z, point.z);
}

bool
bounding_box::inside(const vector &v) const
{
    return ( v.x > min.x && v.y > min.y && v.z > min.z &&
	    v.x < max.x && v.y < max.y && v.z < max.z);
}

bool
bounding_box::empty() const
{
    return (min.x >= max.x || min.y >= max.y || min.z >= max.z);	
}

/* is b associated with this box ? */
bool
bounding_box::inside(const bounding_box &b) const
{
    bounding_box ints(b);
    ints.intersect(*this);
    return !ints.empty();
}

/****************/

void
object::add_to_optilist(Optilist *l)
{
    if (data->add_children_to_optilist == (void*) shape_data::add_children_to_optilist)
	l->add(this);
    else
	data->add_children_to_optilist(l);
}

