/*
   poly.cc -- routines for manipulation of polynomials in one var

   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */

#include "ray.hh"
#include "proto.hh"
#include "extern.hh"
#include "poly.hh"


/*
 * evaluate a polynomial given by coef, ord at the point x. Some libs have a
 * function poly() to do this, but it is not ANSI-C, so this one is
 * included, just to be on the safe side.
 *
 * It's taken from Vort.
 */
flt
Poly::eval(flt x) const
{             
    const flt *fp = coef+ord;
    flt f = *fp;

    /* use Horner's method to evaluate polynomial */
    for (fp--; fp >= coef; fp--)
        f = x * f + *fp;

    return f;
}

void
Poly::multiply(const Poly & p1, const Poly & p2)
{
    Poly dest;

    dest.ord = p1.ord + p2.ord;
    if (dest.ord > MAX_ORDER)
        errormsg("MAX_ORDER exceeded");

    for (int i = 0; i <= dest.ord; i++)
    {
        dest.coef[i] = 0;
        for (int j = 0; j <= i; j++)
            if (i - j <= p2.ord && j <= p1.ord)
                dest.coef[i] += p1.coef[j] * p2.coef[i - j];
    }
    
    *this = dest;
}

void
Poly::differentiate()
{
    flt *fp = coef;
    for (int i = 1; i<= ord; i++) {
	fp[0] = fp[1] * i;
	fp++;
    }
    ord--;
}

void
Poly::power(int exponent, const Poly & src)
{
    int         e = exponent;
    Poly dest(1), base(src);
    
    // classic int power. invariant: src^exponent = dest * src ^ e
    // greetings go out to Lex Bijlsma & Jaap vd Woude
    while (e > 0)
    {
        if (e % 2)
        {
            dest.multiply(dest, base);
            e--;
        } else
        {
            base.multiply(base, base);
            e /= 2;
        }
    }
    *this = dest;
}

void
Poly::clean()
{
    int     i;
    for (i = 0; i <= ord; i++)
    {
        if (abs(coef[i]) < FUDGE)
            coef[i] = 0.0;
    }

    for (i = ord + 1; i <= MAX_ORDER; i++)
        coef[i] = 0;

    while (ord > 0 && coef[ord] == 0.0)
        ord--;
}

void
Poly::real_clean()
{
    int     j;
    flt     d;
    clean();

    /* kill intersections close to zero */
    if (coef[1])
    {
        d = coef[0] / coef[1];

        /* this little fragment of code unearthed a GCC bug! */
        if (d > -((flt) tolerance) && d < ((flt) tolerance))
        {
            /* divide the poly by it's variable t */
            for (j = 0; j < ord; j++)
                coef[j] = coef[j + 1];
            ord--;
        }
    }
}

void
Poly::add(const Poly & p1, const Poly & p2)
{
    int    tempord = (p1.ord < p2.ord) ? p2.ord : p1.ord;

    for (int i = 0; i <= tempord; i++)
    {
     flt temp = 0.0;
        if (i <= p1.ord)
            temp += p1.coef[i];
        if (i <= p2.ord)
            temp += p2.coef[i];
        coef[i] = temp;
    }
    ord = tempord;
}

void
Poly::scalarmultiply(flt fact)
{
    for (    int     i = 0; i <= ord; i++)
        coef[i] *= fact;
}

void
Poly::subtract(const Poly & p1, const Poly & p2)
{
 
    int tempord = (p1.ord < p2.ord) ? p2.ord : p1.ord;
    for (   int     i = 0; i <= tempord; i++)
    {
        flt     temp = 0.0;	// can't store result directly.. a=a-b
        if (i <= p1.ord)
            temp += p1.coef[i];
        if (i <= p2.ord)
	    temp -= p2.coef[i];
	coef[i]         = temp;
    }
    ord = tempord;
}

void
Poly::negate(const Poly & src)
{
    ord = src.ord;

    for (    int     i = 0; i <= src.ord; i++)
        coef[i] = -src.coef[i];
}

/// mod of #u/v#
int
Poly::mod(const Poly &u, const Poly &v)
{
    (*this) = u;
    
    if (v.lc() < 0.0)
    {
        for (int k = u.ord - v.ord - 1; k >= 0; k -= 2)
            coef[k] = -coef[k];

        for (int k = u.ord - v.ord; k >= 0; k--)
            for (int j = v.ord + k - 1; j >= k; j--)
                coef[j] = -coef[j] - coef[v.ord + k] * v.coef[j - k];
    } else {
        for (int k = u.ord - v.ord; k >= 0; k--)
            for (int j = v.ord + k - 1; j >= k; j--)
                coef[j] -= coef[v.ord + k] * v.coef[j - k];
    }

    int k = v.ord - 1;
    while (k >= 0 && coef[k] == 0.0)
        k--;

    ord = (k < 0) ? 0 : k;
#ifdef DEBUG
    debug_clean();
#endif
    return ord;
}

/**
  calculates the modulus of u(x) / v(x) leaving it in *this, it returns 0 if
  r(x) is a constant.
 
  note: this function assumes the leading coefficient of v is 1 or -1
 */
void
Poly::debug_clean()
{
    for (int i = ord+1; i < MAX_ORDER+1; i++)
	coef[i]=0.0;
}
void
Poly::check_sol(flt x) const 
{
    flt f=eval(x);
    Poly p(*this);
    p.differentiate();
    flt d = p.eval(x);
    if( abs(f) > abs(d) * tolerance)
	warning("x=%f is not a root of polynomial\n"
		"f(x)=%f, f'(x)=%f \n", x, f, d);	
}
	
void
Poly::check_sols(const flt*roots , int n) const
{
    for (int i=0; i<n; i++)
	check_sol(roots[i]);
}

void
Poly::print() const
{
#ifdef DEBUG
    int d = ord;
    bool didpr=false; 
    dprintf("poly ord %d: ", d);

    do {	
        if (coef[d])
        {
	    didpr=true;
            dprintf("%f", coef[d]);
            if (d == 1)
                dprintf(" t + ");
            else if (d > 1)
            {
                dprintf(" t^%d + ", d);
            }
        }
    }
    while (d-- > 0);
    if (!didpr)
	dprintf("0.0");
    dprintf("\n");
#endif          /* DEBUG */
}

