/*
   polygon.c -- handle polygon type objects.

   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
*/

#include "stats.hh"
#include "object.hh"
#include "triangle.hh"
#include "vray.hh"
#include "proto.hh"
#include "extern.hh"
#include "ibuf.hh"


polygon_data::~polygon_data()
{
    delete[] vertices;
    delete[] clips;
}


void
polygon_data::scale(vector s)
{
    for (int     i = 0; i < no_vertices; i++)
        vertices[i].scale(s);
}


void
polygon_data::rotate(const matrix & rotmat)
{
    for (int     i = 0; i < no_vertices; i++)
        vertices[i].rotate  (rotmat);
}


void
polygon_data::translate(vector t)
{
    for (int     i = 0; i < no_vertices; i++)
        vertices[i] += t;
}


void
polygon_data::print() const
{
#ifdef DEBUG
    dprintf("projecting along %c %s\n", project_to+'x', (convex) ? ", convex" : "");
    dprintf("%d vertices {\n", no_vertices);
    bmax.print("max");
    bmin.print("min");
    for (int     i = 0; i < no_vertices; i++)
        vertices[i].print("vertex");

    dprintf("}\n");
#endif
}

bool
polygon_data::point_in_concave_polygon(flt t, ray * r) const
{

    int  crosses = 0;
    int    u = (project_to + 1) %3 ;
    int	   v = (project_to + 2 )%3;

    if( u > v) {
	int j = u;
	u = v ;
	v =j;
    }
    vector  x;
    x[project_to]= 0.0;
    x[u] = r->pos[u] + t * r->dir[u];
    x[v] = r->pos[v] + t * r->dir[v];

    /* bounding box test. Inspired on GGems */
    if ((x[u] < bmin[u]) || (x[u] > bmax[u]) || (x[v] < bmin[v]) || (x[v] > bmax[v]))
	return false;

    for (int i = 0; i < no_vertices; i++)
    {
	int    j = (i + 1) % no_vertices;

	int    qi = 0;
	int    qj = 0;
	int     ri = 0;
	int     rj = 0;

        if (vertices[i][v] == vertices[j][v])
            continue;           /* ignore horizontal lines */

        /* If we are both above, or both below the intersection point, go onto the next one. */
        if (vertices[i][v] < x[v])
            qi = 1;
        if (vertices[j][v] < x[v])
            qj = 1;
        if (qi == qj)
            continue;

        /* We know one end point was above, and one was below. If
         * theyare both to the left, then we crossed the line to
         * negative infinity, and we continue.  */

        if (vertices[i][u] < x[u])
            ri = 1;
        if (vertices[j][u] < x[u])
            rj = 1;

        if (ri & rj)
        {
            crosses++;
            continue;
        }
        /* Otherwise, if we are both to the right, we can continue without a cross. */

        if ((ri | rj) == 0)
            continue;

        /* more difficult acceptance... We have a line segment which
          occurs with endpoints in diagonally opposite quadrants.  We
          must solve for the intersection, ie, where v = 0.
	 */

	flt m = (vertices[j][v] - vertices[i][v]) /
            (vertices[j][u] - vertices[i][u]);

	flt b = (vertices[j][v] - x[v]) -
            m * (vertices[j][u] - x[u]);

        if ((-b / m) < EPSILON)
            crosses++;
    }
    return (crosses & 1);
}

bool
polygon_data::point_in_convex_polygon(flt t, ray * r) const
{
    /* clip plane to polygon. */
    int    u = (project_to+1)%3;
    int    v = (project_to+2)%3;
    int    w = (project_to)%3;

    vector x;
    x[w] = 0.0;
    x[u] = r->pos[u] + t * r->dir[u];
    x[v] = r->pos[v] + t * r->dir[v];

    /* bounding box test. Inspired on GGems */
    if ( (x[u] < bmin[u]) || (x[u] > bmax[u]) || (x[v] < bmin[v]) || (x[v] > bmax[v]) )
	return false;
    vector a(0);
    for (int     i = 0; i < no_vertices; i++){
	a[u] = x[u] - vertices[i][u];
	a[v] = x[v] - vertices[i][v];
	
	if (a[u] * clips[i][u] + a[v] * clips[i][v] < 0)
	    return false;
    }


    return true;
}


bool    
polygon_data::all_intersections(ibuf * q, ray * r, int flags, bool &) const
{
    stat->test++;

    flt         d = r->dir * normalvector;
    if (ISZERO(d))
        return false;

    dqueue  inters;
    inters.t = (mov - r->pos * normalvector) / d;
    if (inters.t < tolerance || inters.t > r->maxt)
        return false;


    bool        b = (convex)?point_in_convex_polygon(inters.t, r) : point_in_concave_polygon(inters.t, r);

    if (b)
    {
        stat->hit++;
        inters.obj = daddy;
        q->add_to(flags, inters);
        return true;
    } else
        return false;
}


void
polygon_data::bound(matrix *trans, bounding_box *bbox)
{
    assert(!trans);
    bbox->set_empty();
    for (int i = 0; i < no_vertices; i++) {
	bbox->add(vertices[i]);
    }
    bmax = bbox->max;
    bmin = bbox->min;
}

void
polygon_data::precompute()
{
    stat->howmuch ++;

    if (no_vertices < 3)
    {
        print();
        errormsg("polygon must have at least 3 vertices. ");
    }
    vector *v = vertices;

    vector a = v[1] - v[0];
    vector b = v[2] - v[1];

    normalvector = a ^ b;
    normalvector /= normalvector.length();
    mov = v[0] * normalvector;

    /* project along which axis ? */
    project_to = normalvector.dominant();
    b[project_to] =a[project_to] =0;

    vector proj = a ^ b;

    clips = new vector[no_vertices];
    vector chknormal,
            last = a;
    int     i=0;
    for (; i < no_vertices - 1; i++)
    {
        a = v[i + 1] - v[i];
	last[project_to] = a[project_to] = 0;
	clips[i] = proj ^ a;

        /* check convexity. */
        chknormal = last ^ a;
        last = a;
        if (chknormal * normalvector < 0) {
            convex = false;
	}

        if (!clips[i])
            warning("degenerate polygon");
    }
    if (convex) {
    a = v[0] - v[i];
    a[project_to] = 0;
    clips[i] = proj ^ a;

    chknormal = last ^ a;
    if (chknormal * normalvector < 0)
        convex = false;
    }
    if (!convex)
    {
        delete[]  clips;
        clips = 0;
    }
    prims++;
}


vector  
polygon_data::normal(vector ) const
{
    return normalvector;
}

/* initialize a polygon_data */
shape_stats *polygon_data::stat = shape_stats::regist("polygon");

polygon_data::polygon_data(svec<vector> v)
{
    type = POLYGON;
    bmin.set(-INFTY);
    bmax.set(INFTY);
    normalvector.set(0);
    no_vertices = v.sz();
    mov = 0;
    clips = 0;
    vertices = v.copy_array();
    convex = true;
    project_to = X;
}


polygon_data::polygon_data (polygon_data const & src) {
    bmin = src.bmin; bmax = src.bmax;
    no_vertices = src.no_vertices;
    normalvector = src.normalvector;
    mov = src.mov;              /* mov (see plane_data) */
    convex = src.convex;
    project_to = src.project_to;
    clips=  0;

    vertices = new vector[no_vertices];
    for (int i=  0; i < no_vertices; i ++)
	vertices[i] = src.vertices[i];
}

PUBLIC object *
        get_new_polygon_object(svec<vector> sv)
{
    object *o = new object;
    o->data = new polygon_data(sv);
    o->data->setdaddy(o);
    return o;
}


