/*
   quadric.c --  quadratic shape routines.

   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */


#include "object.hh"
#include "shapes.hh"
#include "proto.hh"
#include "extern.hh"
#include "ibuf.hh"
#include "poly.hh"
#include "matrix.hh"

/* intersect ray with quadratic form. */
int quadric_data::do_intersection(ray * r, flt * roots, int) const
{
    
    flt     a1 = r->dir.x;
    flt     a2 = r->dir.y;
    flt     a3 = r->dir.z;
    flt     b1 = r->pos.x;
    flt     b2 = r->pos.y;
    flt     b3 = r->pos.z;


    /* coefficients of t^2, t^1 and t^0 */
        Poly   quad_poly;
    quad_poly.coef[2] = qt.x * sqr(a1) + qt.y * sqr(a2) + qt.z * sqr(a3)
        + bqt.x * a1 * a2 + bqt.y * a2 * a3 + bqt.z * a3 * a1;

    quad_poly.coef[1] = 2 * (qt.x * a1 * b1 + qt.y * a2 * b2 + qt.z* a3 * b3)
        + bqt.x * (a1 * b2 + b1 * a2)
        + bqt.y * (a2 * b3 + a3 * b2)
        + bqt.z * (a1 * b3 + a3 * b1)
        + lt.x * a1 + lt.y * a2 + lt.z * a3;

    quad_poly.coef[0] = qt.x * sqr(b1) + qt.y * sqr(b2) + qt.z* sqr(b3)
        + bqt.x * b1 * b2 + bqt.y * b2 * b3 + bqt.z * b1 * b3 +
        lt.x * b1 + lt.y * b2 + lt.z * b3 + a;

    quad_poly.ord = 2;
    int         sols = quad_poly.solve_rt( roots, true);      /* solve polynomial. */

    return sols;
}
bool
quadric_data::inside(vector x) const
{
    return (qt.x * sqr(x.x) + qt.y * sqr(x.y) + qt.z * sqr(x.z)
            + bqt.x * x.x * x.y + bqt.y * x.z * x.y + bqt.z * x.x * x.z
            + lt.x * x.x + lt.y * x.y + lt.z * x.z + a < 0);
}

/* normal to a quadric. It's just the gradient at the desired point */
vector  quadric_data::normal(vector x) const
{
    vector  n;
    /* calculate gradient: it's (d/dx =  d/dy *  d/dz , function) */
    n.set(
          2 * qt.x * x.x + bqt.x * x.y + bqt.z * x.z + lt.x,
          2 * qt.y * x.y + bqt.y * x.z + bqt.x * x.x + lt.y,
          2 * qt.z * x.z + bqt.z * x.x + bqt.y * x.y + lt.z);

    return n;
}

void
quadric_data::translate(const vector t)
{
    quadric_data quad = *this;
    /* substitute x:=x-t.x, y:=y-t.y, z:=z-t.z in the quadratic equation. Is there a smarter way of doing this? */

    a += -(quad.lt.x * t.x + quad.lt.y * t.y + quad.lt.z * t.z) +
        quad.bqt.x * t.x * t.y + quad.bqt.y * t.y * t.z + quad.bqt.z * t.z * t.x +
        quad.qt.x * sqr(t.x) + quad.qt.y * sqr(t.y) + quad.qt.z * sqr(t.z);

    lt.x += -2 * t.x * quad.qt.x + -t.y * quad.bqt.x - t.z * quad.bqt.z;
    lt.y += -2 * t.y * quad.qt.y + -t.z * quad.bqt.y - t.x * quad.bqt.x;
    lt.z += -2 * t.z * quad.qt.z + -t.x * quad.bqt.z - t.y * quad.bqt.y;
}
/*
 * rotate quadric: a quadric is determined by it's equation Q: (Ax,x) +
 * (b,x) + c = 0. (A matrix, b vector, c constant) so if x \in RQ (R is a
 * rotation, then R^{-1}x \in Q, so (AR^-1x, x) + (b, R^-1x) + c = 0, or
 * (RAR^Tx,x) + (Rx,b)  +c = 0.
 */

void
quadric_data::rotate(const matrix & rotmat)
{
    matrix
        prod,
        transm;
    vector  lin;

    matrix  quad(0);
    quad[0][0] = qt.x;
    quad[1][1] = qt.y;
    quad[2][2] = qt.z;
    quad[0][1] = quad[1][0] = bqt.x / 2.0;
    quad[1][2] = quad[2][1] = bqt.y / 2.0;
    quad[0][2] = quad[2][0] = bqt.z / 2.0;

    transm = rotmat.transpose();

    /* calc AR^T, store result in prod, quad is no longer needed */
    prod = quad * transm;

    /* calc RAR^T, store result in quad */
    quad = rotmat * prod;

    qt.x = quad[0][0];
    qt.y = quad[1][1];
    qt.z = quad[2][2];

    bqt.x = quad[0][1] + quad[1][0];
    bqt.y = quad[1][2] + quad[2][1];
    bqt.z = quad[0][2] + quad[2][0];

    lin.x = lt.x;
    lin.y = lt.y;
    lin.z = lt.z;
    lin.rotate(rotmat);
    lt.x = lin.x;
    lt.y = lin.y;
    lt.z = lin.z;
}

void
quadric_data::scale(const vector s)
{
    flt     A,
            b,
            c;
    A = 1 / s.x;
    b = 1 / s.y;
    c = 1 / s.z;

    qt.x *= sqr(A);
    qt.y *= sqr(b);
    qt.z *= sqr(c);

    bqt.x *= A * b;
    bqt.y *= b * c;
    bqt.z *= c * A;

    lt.x *= A;
    lt.y *= b;
    lt.z *= c;
}

#define termprint(a,b) if (a) dprintf("%5lf" #b " + ", a)
void
quadric_data::print() const
{
#ifdef DEBUG
    termprint(qt.x, x^2);
    termprint(qt.y, y^2);
    termprint(qt.z, z^2);

    termprint(bqt.x, xy);
    termprint(bqt.y, yz);
    termprint(bqt.z, zx);

    termprint(lt.x,x);
    termprint(lt.y,y);
    termprint(lt.z,z);
    dprintf("%5lf = 0\n",a);
#endif          /* DEBUG */
}

shape_stats *quadric_data::stat = shape_stats::regist("quadric");
quadric_data::quadric_data(vector a1, vector a2, vector a3, flt a4)
{
    type = QUADRIC;
    qt = a1;
    bqt = a2;
    lt =a3;
    a = a4;
}


void
quadric_data::precompute()
{
    stat->howmuch++;
    prims++;
}


PUBLIC object *
        get_new_quadric_object(vector ia,vector ib,vector ic,flt r)
{
    object *o = new object;
    o->data = new quadric_data(ia,ib,ic,r);
    o->data->setdaddy(o);
    o->flags |= O_INFINITE | O_CLOSED ;

    return o;
}
