/*
  solve.cc -- solve equations
 
  The solve_quartic and
  solve_cubic are from Graphics Gems I "roots3and4.c", by Jochen
  Schwarze.
 */
#include "ray.hh"
#include "proto.hh"
#include "extern.hh"
#include "poly.hh"

/* cubic root. */
inline flt cbrt(flt x)
{
    if (x > 0.0)
	return pow(x, 1.0/3.0) ;
    else if (x < 0.0)
	return -pow(-x, 1.0/3.0);
    else
	return 0.0;
}

static int     solve_cubic(const flt * coeffs, flt * roots);
static int     solve_rt_linear(const flt * coeff, flt * roots);
static int     solve_rt_quartic(const flt * coef, flt * roots);
static int     solve_rt_cubic(const flt * coeffs, flt * roots);
int             solve_rt_quadric(const flt * coeffs, flt * roots);

// arght.
int (*solve_table[])(const flt *, flt *) = {
  0,
  solve_rt_linear,
  solve_rt_quadric,
  solve_rt_cubic,
  solve_rt_quartic,
};

/***************************************************************************
   ENTRY POINT: solve polynomial, return number of nonnegative roots       *
 ***************************************************************************/

int
Poly::solve_rt(flt *roots, bool chkall)
{
    /* avoid degenerate cases. */
    if (ord >= 3) {
	/* especially avoid intersections close to 0 */
	real_clean();
    } else {
	clean();
    }
    
    int             n;	
    if (ord > 4)
	n = sturm_solve_rt(roots, chkall);
    else    /* special cases can be handled much faster: */
	n = solve_table[ord](coef, roots);
    

#ifdef DEBUG
    assert(n>=0);
    if (debug_options & DEBUGSELFCHECK)
	    check_sols(roots,n);
#endif

    return n;
}



/* solve linear equation */
static int
solve_rt_linear(const flt * coeff, flt * roots)
{
    if ((*roots = -coeff[0] / coeff[1]) > tolerance)
	return 1;
    else
	return 0;
}

/*
   SPECIAL ROOTFINDERS: quadric, cubic and quartic.
 *
   written by Jochen Schwarze and me from Graphics Gems I, by Jochen
   Schwarze, (schwarze@isa.de) quadratic: both by Jochen Schwarze,
   from graphics gems cubic: one from me, and one by Jochen Schwarze
   quartic: by Jochen Schwarze.
 
 
   Here is the original comment header:
 
   Roots3And4.c
 
   Utility functions to find cubic and quartic roots, coefficients are passed
   like this:

   c[0] + c[1]*x + c[2]*x^2 + c[3]*x^3 + c[4]*x^4 = 0

   The functions return the number of non-complex roots and put the values
   into the s array.

   Author:         Jochen Schwarze (schwarze@isa.de)

   Jan 26, 1990 Version for Graphics Gems Oct 11, 1990 Fixed sign
   problem for negative q's in SolveQuartic (reported by Mark
   Podlipec), Old-style function definitions, IsZero() as a macro Nov
   23, 1990 Some systems do not declare acos() and cbrt() in <math.h>,
   though the functions exist in the library. If large coefficients
   are used, EQN_EPS should be reduced considerably (e.g. to 1E-30),
   results will be correct but multiple roots might be reported more
   than once. */




/*
   all roots of quadratic eqn.
 */
int
solve_quadric(const flt c[3], flt s[2])
{      
    /* normal form: x^2 + px + q = 0 */
    flt                p = c[1] / (2 * c[2]);
    flt    q = c[0] / c[2];

    flt D = p * p - q;

    if (D < 0) {
	return 0;
    } else {
	D = sqrt(D);

	s[0] = D - p;
	s[1] = -D - p;
	return 2;
    }
}

/*
   solve quadric, return roots sorted and bigger than 0 double roots are
   returned doubly, so CSG still works.
 */
int
solve_rt_quadric(const flt * coeffs, flt * roots)
{

    /*
     * pre: coeffs[2] != 0
     */
    flt                p = coeffs[1] / (2 * coeffs[2]);
    flt    q = coeffs[0] / coeffs[2];

flt    D = p * p - q;

    if (D < 0) {
	return 0;
    }
    /*
     * don't check for double roots. If there are any, we still need both of them.
     */

    D = sqrt(D);

    flt    r1 = D - p;
    flt    r2 = -D - p;

    if (r1 > r2) {
	D = r2;
	r2 = r1;
	r1 = D;
    }
    /* r2 is maximum. If r2 < epsilon, then shake it */
    if (r2 < tolerance)
	return 0;

    /* r2 > tolerance, but is r1 > tolerance? */
    if (r1 < tolerance) {
	*roots = r2;
	return 1;
    } else {
	*roots++ = r1;
	*roots = r2;
	return 2;
    }

}


/*
   this one is for RT applications, it returns number of roots bigger than
   EPSILON, and the roots itself sorted in an array.
 */

/*
   solve cubic, same method as above. Don't sort roots; return all roots.
 */
static int
solve_cubic(const flt c[4], flt s[3])
{
    /* normal form: x^3 + Ax^2 + Bx + C = 0 */

    flt    A = c[2] / c[3];
    flt    B = c[1] / c[3];
    flt    C = c[0] / c[3];

    /*
     * substitute x = y - A/3 to eliminate quadric term: x^3 +px + q = 0
     */

    flt    sq_A = A * A;
    flt    p = 1.0 / 3 * (-1.0 / 3 * sq_A + B);
    flt    q = 1.0 / 2 * (2.0 / 27 * A * sq_A - 1.0 / 3 * A * B + C);

    /* use Cardano's formula */

    flt    cb_p = p * p * p;
    flt    D = q * q + cb_p;
    int num;
    if (ISZERO(D)) {
	if (ISZERO(q)) {	/* one triple solution */
	    s[0] = 0;
	    num = 1;
	} else {		/* one single and one flt solution */
	    flt            u = cbrt(-q);

	    s[0] = 2 * u;
	    s[1] = -u;
	    num = 2;
	}
    } else if (D < 0) {		/* Casus irreducibilis: three real solutions */
	flt            phi = 1.0 / 3 * acos(-q / sqrt(-cb_p));
	flt            t = 2 * sqrt(-p);

	s[0] = t * cos(phi);
	s[1] = -t * cos(phi + M_PI / 3);
	s[2] = -t * cos(phi - M_PI / 3);
	num = 3;
    } else {			/* one real solution */
	flt            sqrt_D = sqrt(D);
	flt            u = cbrt(sqrt_D - q);
	flt            v = -cbrt(sqrt_D + q);

	s[0] = u + v;
	num = 1;
    }

    /* resubstitute */

    flt                sub = 1.0 / 3 * A;

    for (int i = 0; i < num; ++i)
	s[i] -= sub;

    return num;
}

/*
   solve a quartic eqn. Ferrari was the first to think of this. (I believe,
   or was it Descartes?)
 */

static int
solve_quartic(const flt c[5], flt s[4])
{    
    /* normal form: x^4 + Ax^3 + Bx^2 + Cx + D = 0 */

    flt    A = c[3] / c[4];
    flt    B = c[2] / c[4];
    flt    C = c[1] / c[4];
    flt    D = c[0] / c[4];

    /*
     * substitute x = y - A/4 to eliminate cubic term: x^4 + px^2 + qx + r = 0
     */
    flt    sq_A = A * A;
    flt    p = -3.0 / 8 * sq_A + B;
    flt    q = 1.0 / 8 * sq_A * A - 1.0 / 2 * A * B + C;
    flt    r = -3.0 / 256 * sq_A * sq_A + 1.0 / 16 * sq_A * B - 1.0 / 4 * A * C + D;
    int num;
    if (ISZERO(r)) {
	/* no absolute term: y(y^3 + py + q) = 0 */
	flt            coeffs[4];
	coeffs[0] = q;
	coeffs[1] = p;
	coeffs[2] = 0;
	coeffs[3] = 1;

	num = solve_cubic(coeffs, s);

	s[num++] = 0;
    } else {
	/* solve the resolvent cubic ... */
	flt            coeffs[4];
	coeffs[0] = 1.0 / 2 * r * p - 1.0 / 8 * q * q;
	coeffs[1] = -r;
	coeffs[2] = -1.0 / 2 * p;
	coeffs[3] = 1;

	(void) solve_cubic(coeffs, s);

	/* ... and take the one real solution ... */

	flt	z = s[0];

	/* ... to build two quadric equations */

	flt	u = z * z - r;
	flt v = 2 * z - p;

	if (ISZERO(u))
	    u = 0;
	else if (u > 0)
	    u = sqrt(u);
	else
	    return 0;

	if (ISZERO(v))
	    v = 0;
	else if (v > 0)
	    v = sqrt(v);
	else
	    return 0;

	coeffs[0] = z - u;
	coeffs[1] = q < 0 ? -v : v;
	coeffs[2] = 1;

	num = solve_quadric(coeffs, s);

	coeffs[0] = z + u;
	coeffs[1] = q < 0 ? v : -v;
	coeffs[2] = 1;

	num += solve_quadric(coeffs, s + num);
    }

    /* resubstitute */
    flt sub = 1.0 / 4 * A;

    for (int i = 0; i < num; ++i)
	s[i] -= sub;

    return num;
}


static int
dcomp(const void *p, const void *q)
{
    flt a = *(flt *) p;
    flt   b = *(flt *) q;
    if (a < b)
	return -1;
    if (a > b)
	return 1;
    return 0;
}

static int
solve_rt_quartic(const flt * coef, flt * roots)
{
    flt            r[4];
    int                n = solve_quartic(coef, r);
    qsort(r, n, sizeof(flt), dcomp);

    int j = 0;
    for (int i = 0; i < n; i++)
	if (r[i] > tolerance)
	    roots[j++] = r[i];
    return j;
}

static int
solve_rt_cubic(const flt * coeffs, flt * roots)
{
    flt            r[4];
    int                 n = solve_cubic(coeffs, r);
    qsort(r, n, sizeof(flt), dcomp);

    int     j = 0;
    for (int             i = 0; i < n; i++)
	if (r[i] > tolerance)
	    roots[j++] = r[i];
    return j;
}

/*
   the debugging routines are still here, if you need them.
 */


/* print solutions, and check wether they are solutions */
void
Poly::print_sols(flt const* s, int nosols) const
{
#ifdef DEBUG
    if (nosols <= 0) {
	printf("no solutions\n");
	return;
    }
    printf("%d solution(s)\n", nosols);

    for (    int             i = 0; i < nosols; i++)
	printf("%d: p(%f) = %f\n", i + 1, s[i], eval(s[i]));

    printf("\n");
#endif				/* DEBUG */
}
