/*
  texture.cc --  manipulations for textures

  (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>
  It is based on "Ray" by George Kyriazis <kyriazis@turing.cs.rpi.edu>
  Rayce is released under the terms of the GNU General Public License,
  it has no warranty what so ever. For more details: see the file "COPYING"
 */


#include "texture.hh"
#include "matrix.hh"
#include "proto.hh"
#include "extern.hh"
#include "ibuf.hh"

/****************************************************************/
void
local_texture::init(local_finish const *f)
{
    if (f)
	* (local_finish*) this = *f;
    else { // pov defaults
	local_finish initter;
	
	* (local_finish*) this = initter;
	
	diffuse.set(0.6);
	ambient.set(0.2);
	specular.set(0);
	refl_color.set(0);
	refr_color.set(0);
    }
}

/****************************************************************/
simple_texture::simple_texture(){
    normal = 0;
    pigment = 0;
    finish = 0;
}

simple_texture::~simple_texture(){
    delete normal;
    delete pigment;
    delete finish;
}


local_texture
simple_texture::get_local_texture (vector x, vector n) const
{
    local_texture t(n);
    t.init(finish);

    if (pigment) {
	color c = pigment->get_color(x,n);
	
	t.ambient = t.ambient.length() * c;
	t.diffuse = t.diffuse.length() * c;

	/* specular always looks white in PoV */
	color orgspec  = t.specular;
	t.specular.set(1);
	t.specular *= orgspec.length();       
	t.refl_color = t.refl_color.length() * c;
	t.refr_color = t.refr_color.length() * c;
    }

    if (normal)
	t.normal += normal->get_delta_normal();
    
    return t;
}

/* copy texture */
simple_texture::simple_texture(simple_texture & src)
{    
    pigment  = (src.pigment)  ? src.pigment->clone() : 0;
    normal   = (src.normal)   ? src.normal->clone() : 0;
    finish   = (src.finish)   ? new local_finish( *src.finish) : 0;
}

void
simple_texture::translate(vector v)
{
    if (normal) normal->translate (v);
    if (pigment) pigment->translate (v);
}


void
simple_texture::rotate(const matrix & rotmat)
{
    if (normal) normal->rotate (rotmat);
    if (pigment)pigment-> rotate (rotmat);
}


void
simple_texture::scale(const vector s)
{
    if (normal) normal->scale (s);
    if (pigment) pigment->scale (s);
}


void
simple_texture::print() const
{
#ifdef DEBUG
    dprintf("simpletexture {\n");
    if (normal)
	normal->print();
    if (pigment)
	pigment->print();
    if(finish)
	finish->print();
    dprintf("}\n");
#endif          /* DEBUG */
}



/****************************************************************/
void pov_compatibility::_print() const {
    dprintf("POV comp ");
    c.print();
}

color pov_compatibility::get_color(vector , vector ) const {
    return c;
}


/****************************************************************/
pigment_modifier::pigment_modifier(pigment_modifier &src)
{
    inv_trans = (src.inv_trans)? new matrix(*src.inv_trans):0;
}

pigment_modifier::~pigment_modifier()
{
    delete inv_trans;
}
pigment_modifier::pigment_modifier()
{
    inv_trans = 0;
}

void
pigment_modifier::rotate(matrix const &rm)
{
    if (!inv_trans)
        inv_trans = new matrix(1);
    *inv_trans = *inv_trans * rm.transpose();
}

void
pigment_modifier::translate(vector t)
{
    if (!inv_trans)
        inv_trans = new matrix(1);

    matrix  m;
    m.inv_translate(t);
    *inv_trans = *inv_trans * m;
}

void
pigment_modifier::print () const
{
    dprintf("pigment { \n");
    if (inv_trans)
	inv_trans->print();
    _print();
    dprintf("}\n");
}

void
pigment_modifier::scale (vector s)
{
    if (!inv_trans)
        inv_trans = new matrix(1);
    matrix  m;
    m.inv_scale(s);
    *inv_trans = *inv_trans * m;
}

