/*
  texture.hh --  texture class hierarchy

  Rayce is (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>
  It is based on "Ray" by George Kyriazis <kyriazis@turing.cs.rpi.edu>
  Rayce is released under the terms of the GNU General Public License,
  it has no warranty what so ever. For more details: see the file "COPYING"
 */

#ifndef TEXTURE_HH
#define TEXTURE_HH


#if 0
/* this is a comment */
// hierarchy:

struct normal_modifier;

/* pigments: pov compatibility, colormaps, imagemaps etc */
/* give ONE color for each location */

class pigment_modifier;
class pov_compatibility : pigment_modifier;
class image_map : pigment_modifier;


class local_finish;
struct local_texture  : local_finish;

struct texture;
class simple_texture  : texture;

class pixel_data
struct colmapped_data : pixel_data;
struct truecol_data   : pixel_data;

pixel_data *search_pixels(char *fn);
void add_pixels(pixel_data *);
#endif

#include "color.hh"
#include "ibuf.hh"

enum
{                               /* texturing. */
    T_EGAL,                     /* smooth */
    T_BLEND,                    /* texture blend (not yet) */
    T_IMAGEMAP,                 /* image mapping */
    T_COLORMAP,                 /* color mapped texture: perlin style (not yet) */
};

enum  maptype{
    M_PLANE=0,
    M_SPHERE=1,
    M_CYLINDER=2,
    M_TORUS=5,
};

enum distributions
{
    DPHONG,
    DBLINN,
    DGAUSS,
    DCOOK,
    DREITZ,
};

enum geometric
{
    GEO_NONE,
    GEO_TORRANCE,
    GEO_SANCER
};
/****************************************************************/

/// bump mapping, bump functions etc.
struct normal_modifier {
    virtual vector get_delta_normal() = 0;
    virtual void print() const = 0;
    virtual void rotate(matrix const & m) =0;
    virtual void scale(vector s)=0;
    virtual void translate(vector v)=0;
    virtual normal_modifier *clone()=0;   
};

/****************************************************************/


/// give ONE color for each location
class pigment_modifier {
    
    matrix *inv_trans;
    
public:
    pigment_modifier(pigment_modifier & src);
    virtual color get_color(vector x, vector n) const = 0;
    virtual void _print() const = 0;
    void print() const;
    pigment_modifier();
    virtual ~pigment_modifier();
    virtual pigment_modifier *clone() = 0;

    virtual void rotate(matrix const& m);
    virtual void scale(vector s);
    virtual void translate(vector v);
};

/** pigments: pov compatibility, colormaps, imagemaps etc */


/// texture has one color in pigment block
class pov_compatibility : public pigment_modifier {
    color c;
    pov_compatibility(pov_compatibility&s) : c(s.c) { }
 public:

    pov_compatibility(color k ) { c =k; }
    color get_color(vector x,vector n) const;
    void _print() const;
    pigment_modifier *clone() { return new pov_compatibility(*this); }
};


/// each point has color from a pixel-file
class image_map : public pigment_modifier {
    /// the picture 
       
    pixel_data *pic;
    /// inverse mapping
    void (*inv_map) (flt &u, flt & v, vector x);
    /** translate a 3D point to a 2D coordinate system */

    maptype type;
public:
    
    /// only one time? 
    bool    once; 

    /// not used 
    bool    interpolated;
    bool    usenormal;

    /// swap output of inv_map? 
    bool uvswap;      

    /// translate image over surface 
    flt     u_offs,
            v_offs;

    /// scale it. 
    flt     u_range,
            v_range;           
    

    pigment_modifier *clone() { return new image_map(*this); }
    image_map(pixel_data *p);
    ~image_map();
    
    void _print() const ;
    void        set_type(int);
    color	get_color(vector x, vector n) const;
 
};


/****************************************************************/

/// define surface characteristics for one point (color, reflectivity, etc.)
class local_finish {

    distributions dtype;
    geometric gtype;

    flt (*set_rough_func)(flt);
    
 public:
    color   refl_color,         /* reflective color */
            refr_color,         /* refractive color */
            ambient,            /* ambient color */
            diffuse,            /* diffuse color */
            specular;           /* specular color */

    flt     roughness;          /* specular roughness: 0 < roughn <=1 */
    flt     index;              /* index of refraction */
    flt     refl_diffuse;       /* circle of diffusion in reflection */
    flt     refr_diffuse;       /* circle of diffusion in refraction */
    flt     brilliance;         /* tightness of diffuse illum. */

    void set_distr(int);
    void set_geo(int);


    void set_rough(flt);
    flt (*distfunc)(flt,vector,vector,vector);
    flt (*geofunc)(vector,vector,vector);
    local_finish();
    virtual ~local_finish() {  }
    void print() const;
};

/****************************************************************/

/// the normal and  finish in some point 
struct local_texture : local_finish 
{
    local_texture(vector & r) 
         { normal = r; }
    void init(local_finish const *t=0 ); // sorry
	
    vector normal;
};

/// has a  normal and a   finish in each  point 
struct texture {
    virtual local_texture get_local_texture (vector,vector) const = 0;
    virtual ~texture() {};
    virtual void print() const =0;
    virtual texture  * clone ()  = 0;
    virtual    void rotate(matrix const& m)=0;
    virtual    void scale(vector s)=0;
    virtual    void translate(vector v)=0;    
};

/// simple texture: no nested textures 
class simple_texture : public texture {    
    matrix *inv_trans;
    
    void operator = (int) { 	} // not used for now
public:

    normal_modifier *normal;
    pigment_modifier *pigment;
    local_finish *finish;

    simple_texture(simple_texture& src);
    simple_texture();
    ~simple_texture () ;
    local_texture get_local_texture(vector,vector) const;
    void print() const;

    void rotate(matrix const& m);
    void scale(vector s);
    void translate(vector v);
    
    texture *clone ()  { return new simple_texture(*this); }
};

#endif


