/*
   trace.c -- fire eyerays, sample rays, pixel output.

   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */

#include	"ray.hh"
#include	"proto.hh"
#include	"extern.hh"
#include 	"ibuf.hh"
#include 	"color.hh"
#include 	"camera.hh"
#include 	"object.hh"
#include 	"grid.hh"
#include 	"light.hh"
#include	"stats.hh"

#ifdef __GNUC__
/* Gcc needs this for SEEK_EOF */
#include	<unistd.h>
#endif

class Pixel_output
{
private:
    FILE *fp;

public:
    int     nextx,
            nexty;              /* where we would start tracing */
    int     xres,
            yres;               /* output file resolution */
            Pixel_output(const char *fname, bool);
            ~Pixel_output();
    void write(byte_color c);
};


PRIVATE void line_stat_update();

static int cur_x,
        cur_y;                  /* current x,y of pixels we're tracing */

/*
   sample a ray inside specified solid angle

   PARAMS:

   r:		ray to modified.
   filter_factor:
   theta:	angle (rad)
   g:		contains info on grid/randomization
 */
void
sample_ray(ray * r, flt * filter_factor, flt theta, Grid_tree * g)
{
    flt     randfact1,
            randfact2;
    g->sample(randfact1, randfact2, *filter_factor, r->type);

    if (ISZERO(theta))          /* don't go thru difficult stuff if you don't have to */
        return ;

    vector  tmp(1, 0, 0);
    vector  hor = tmp ^ r->dir;
    if (tmp.quicklength() < EPSILON)
    {                           /* tmp x r.dir is nullvector */
        tmp.set(0, 1, 0);       /* then this won't give nullvector */
        hor = tmp ^ r->dir;
    }
    /* hor is orthogonal to r->dir */
    hor /= hor.length();
    vector  ver = r->dir ^ hor;
    randfact1 *= theta;
    randfact2 *= theta;

    hor *= randfact1;
    ver *= randfact2;

    r->dir += hor + ver;
    r->dir /= r->dir.length();
}



/*
   trace a single straight ray, calculate color.

   PARAMS:

   r: 		ray to be traced.
   g:	   	grid_tree, containing info about sampling, recursion level
   importance:	contribution to endcolor
 */

color
Scene::trace(ray & r, Grid_tree * g, flt importance)
{
    if (g->recursion > ray::line.maxdepth)
        ray::line.maxdepth = g->recursion;

    /* check for intersection with the object */
    intersection inter = all_intersections(r);

    /* if no intersection, return some background color */
    color   col = (!inter.q_ent.obj) ?
    
    bgcolor(&r) :
    
    /* else calculate the shading function there */
    col = shade(&inter, g, importance);


    col += fog_data::get_color(&r, inter.q_ent.t);
    return col;
}

/***************************************************************************
 *  entry point of trace.c.  raytrace the whole scene	   		   *
 ***************************************************************************/

void
Scene::render_to(const char *fname)
{
    shape_stats::start();

    trace_start = true;
    the_output = new Pixel_output(fname, continue_flag);
    xres = the_output->xres;
    yres = the_output->yres;

    flt     sample_factor = 1 / flt(tries);


    cur_x = the_output->nextx;
    for (cur_y = the_output->nexty; cur_y < yres; cur_y++)
    {
        if (cur_x == xres)
            cur_x = 0;
        for (; cur_x < xres; cur_x++)
        {
            color pixcol(0);      /* color gathered sofar for a pixel */

            /* the time blur has to be done in linear time and not randomly */
            int i = tries;

            while (i--)
            {
                check_abort();  /* somebody pressed a key? */


                ray     eyeray(R_EYE,pick_time(i));

		flt     filter_factor;
		the_cam->set_eyeray(eyeray, filter_factor, cur_x, cur_y, xres, yres);

                // get the color and sum all the intensities together */
                pixcol += filter_factor * sample_factor * trace(eyeray, the_grid->tree, 1.0);		
            }
            // normalise color
            the_output->write(pixcol);  // output one pixel */
        }
        line_stat_update();     // and print stats */
    }
    delete the_output;
    shape_stats::print_all();   // global statistics. */
}

/*
   print statistics after abortion

   PARAMS:

   needed for keeping signal() from complaining

*/
void
abort_trace(int)
{
    if (trace_start)
    {
        if (!silent_mode)
        {
            delete  the_scene.the_output;
            line_stat_update();
        }
        shape_stats::print_all();
    } else {
        printf("(interrupt)\n");
    }

    exit(0);	// use exit.. output file needs to be synced.
}
/***************************************************************************
 * statistics								   *
 ***************************************************************************/

/* statistics for a line */
PRIVATE void
line_stat_update()
{
    if (cur_x == 0)
        return ;
    if (cur_x != xres)
        printf("\nStopped at pixel %d\n", cur_x);
    if (!silent_mode)  {
        printf("line %3d. ", cur_y);
	ray::reset_stats();
    }
}
/***************************************************************************
 * FILE OUTPUT								   *
 ***************************************************************************/
const   OUTBUFSIZ = 100;

static void
write_TGA_head(FILE * f, int xres, int yres)
{
    int     header[18];
    for (int i = 0; i < 18; i++)
        header[i] = 0;

    /* put ID and screen res in header */
    header[2] = 0x02;
    header[12] = xres & 0x00ff;
    header[13] = xres / 256;
    header[14] = yres & 0x00ff;
    header[15] = yres / 256;
    header[16] = 24;
    header[17] = 32;

    /* dump header */
    for (int i = 0; i < 18; i++)
        putc(header[i], f);
}

Pixel_output::~Pixel_output()
{
    fclose(fp);
}

/*
   calc the integer value to be put to the file
   and store it. NB targa is in bgr format
 */

void
Pixel_output::write(byte_color col)
{
    fputc(col.b,fp);
    fputc(col.g,fp);
    fputc(col.r,fp);
    if (debug_options & DEBUGRUNTIME)
	fflush(fp);
}

Pixel_output::Pixel_output(const char *fname, bool continuetrace)
{
    nextx = nexty = 0;
    xres = ::xres;
    yres = ::yres;

    if (continuetrace)
    {
        fp = open_file(fname, "r+b", false);
        if (fp) {
            read_TGA_head(fp, xres, yres);
            fseek(fp, 0, SEEK_END);
            long    len = ftell(fp) - 18;
	    if (len % 3)
		errormsg("colors come in three");
	    len /= 3;
            nexty = len / xres;
            nextx = len % xres;
            printf("xres %d x yres %d, starting at line %d, pixel %d\n", xres, yres, nexty, nextx);

            return ;
        } else
            printf("error opening continue file, creating a new file\n");
    }

    fp = open_file(fname, "wb");
    write_TGA_head(fp, xres, yres);
    printf("xres %d x yres %d\n", xres, yres);
}


